/***************************************************************************
                          sqlguidoc_mysql.cpp
                          -------------------
    begin                : 08/02 2002
    copyright            : (C) 2002 by mb
    email                : marcus@beranek.de
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


/*
  This class implements all necessary member-functions of sqlguidoc, which
  must be reimplenented for a special database (here MySQL).
*/

// include files for Qt
#include <qdir.h>
#include <qwidget.h>
#include <qstring.h>
#include <qapplication.h>
#include <qregexp.h>

// include files for KDE
#include <kapplication.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kio/job.h>

// application specific includes
#include "sqlguidoc_mysql.h"
//#include "sqlgui_part.h"
//#include "sqlguiview.h"
#include <qcursor.h>
#include <qapplication.h>

#include "dbtypes.h"

#include <iostream>
#include <iomanip>
using namespace std;

SqlGuiDocMysql::SqlGuiDocMysql(QWidget *parent, const char *name) : SqlGuiDoc(parent, name){
  // nothing to do.
}

SqlGuiDocMysql::~SqlGuiDocMysql(){

}

QString SqlGuiDocMysql::getProtocol(){
  QString p = "mysql";
  return p;
}

int SqlGuiDocMysql::getDBType(){
  return SQLGUI_MYSQL;
}


void SqlGuiDocMysql::slotDbCreate( QString* db ){
  
  if(db->isEmpty()) return;
  // std::cout << "DB-Creating-Request..." << endl;
  // std::cout << db->latin1() << endl;
  
  QString * q = new QString("CREATE DATABASE ");
  q->append( *db );
  
  MYSQL datenbank;
  mysql_init( &datenbank );
  MYSQL_RES * res;
  
  if(!mysql_real_connect( &datenbank, doc_host.latin1() , doc_uname.latin1() , doc_pass.latin1(), "", doc_port, NULL, 0)){
    // std::cout << "Connection failed..." << endl;
    QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
    
    doc_host.setLatin1("");
    doc_uname.setLatin1("");
    doc_pass.setLatin1("");

    emit( DbQueryError( errmsg) );
    return;
  }
  
  int okay = mysql_query( &datenbank, q->latin1() );

  if(okay==0){
    emit( NewDbCreated() );
  } else {
    QString *errmsg = new QString( kl->translate( i18n("Error creating database. Maybe no permission?")));
    emit( DbQueryError( errmsg ) );
  }
  mysql_close( &datenbank );
}


void SqlGuiDocMysql::slotCountRows( QString* _db, QString* _tb){
  QApplication::setOverrideCursor( Qt::waitCursor );

  // std::cout << "Got Count-Row-Request..." << endl;

  if(_db->isEmpty()){
    QApplication::restoreOverrideCursor();
    return;
  }
  if(_tb->isEmpty()){
    QApplication::restoreOverrideCursor();
    return;
  }

  QString * q = new QString("SELECT count(*) AS ROWS FROM ");
  q->append( *_db );
  q->append( "." );
  q->append( *_tb );

  slotDoQuery( _db, q);

  QApplication::restoreOverrideCursor();

}

void SqlGuiDocMysql::slotDbDelete( QString* db ){

  if(db->isEmpty()) return;

  // std::cout << "DB-Deletion-Request..." << endl;

  // std::cout << db->latin1() << endl;

  QString * q = new QString("DROP DATABASE ");
  q->append( *db );
  
  MYSQL datenbank;
  mysql_init( &datenbank );
  MYSQL_RES * res;
  
  if(!mysql_real_connect( &datenbank, doc_host.latin1() , doc_uname.latin1() , doc_pass.latin1(), "", doc_port, NULL, 0)){
    // std::cout << "Connection failed..." << endl;
    QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
    
    doc_host.setLatin1("");
    doc_uname.setLatin1("");
    doc_pass.setLatin1("");
    
    emit( DbQueryError( errmsg) );
    return;
  }

  int okay = mysql_query( &datenbank, q->latin1() );
  
  if(okay==0){
    emit( DbErased() );
  } else {
    QString *errmsg = new QString( kl->translate( i18n("Error removing database. Maybe no permission?")));
    emit( DbQueryError( errmsg ) );
  }

  mysql_close( &datenbank );
}

void SqlGuiDocMysql::slotTableDelete( QString* db, QString* tb ){

  // std::cout << "Got Table-Deletion-Request..." << endl;
  // std::cout << db->latin1() << " " << tb->latin1() << endl;
  if(db->isEmpty()) return;
  if(tb->isEmpty()) return;

  QString * q = new QString("DROP TABLE ");
  q->append( *db );
  q->append( "." );
  q->append( *tb );
  
  MYSQL datenbank;
  mysql_init( &datenbank );
  MYSQL_RES * res;

  if(!mysql_real_connect( &datenbank, doc_host.latin1() , doc_uname.latin1() , doc_pass.latin1(), "", doc_port, NULL, 0)){
    // std::cout << "Connection failed..." << endl;
    QString *errmsg = new QString( kl->translate( i18n("Error connecting to server."))) ;

    doc_host.setLatin1("");
    doc_uname.setLatin1("");
    doc_pass.setLatin1("");

    emit( DbQueryError( errmsg) );
    return;
  }

  int okay = mysql_query( &datenbank, q->latin1() );
  
  if(okay==0){
    emit( TableErased() );
  } else {
    QString *errmsg = new QString( kl->translate( i18n("Error removing table. Maybe no permission?")));
    emit( DbQueryError( errmsg ) );
  }

  mysql_close( &datenbank );
}


void SqlGuiDocMysql::slotServerStatus( ){

  // std::cout << "Got Staus-Request..." << endl;

  MYSQL datenbank;
  mysql_init( &datenbank );

  if(!mysql_real_connect( &datenbank, doc_host.latin1() , doc_uname.latin1() , doc_pass.latin1(), "", doc_port, NULL, 0)){
    // std::cout << "Connection failed..." << endl;
    QString *errmsg = new QString( kl->translate( i18n("Error connecting to server."))) ;

    doc_host.setLatin1("");
    doc_uname.setLatin1("");
    doc_pass.setLatin1("");

    emit( DbQueryError( errmsg) );
    return;
  }
  
  const char * stat = mysql_stat( &datenbank );
  if(stat){	

    QString *s = new QString(stat);
    // z.B.: Uptime: 9607  Threads: 1  Questions: 1185  Slow queries: 0  Opens: 473  Flush tables: 1  Open tables: 64 Queries per second avg: 0.123
    s->replace(QRegExp("Threads"), "\nThreads");
    s->replace(QRegExp("Questions"), "\nQuestions");
    s->replace(QRegExp("Slow"), "\nSlow");
    s->replace(QRegExp("Open"), "\nOpen");
    s->replace(QRegExp("Flush"), "\nFlush");
    s->replace(QRegExp("Queries"), "\nQueries");
    emit serverStatus( QString::QString( *s ) );

    //    emit serverStatus( QString::QString( stat ) );
  } else {
    QString *errmsg = new QString( kl->translate( i18n("Couldn't get server-status."))) ;
    emit( DbQueryError( errmsg) );
  }
  
  mysql_close( &datenbank );
}


void SqlGuiDocMysql::slotTableCheck( QString* db, QString* tb ){

  QApplication::setOverrideCursor( Qt::waitCursor );

  // std::cout << "Got Table-Check-Request..." << endl;

  if(db->isEmpty()){
    QApplication::restoreOverrideCursor();
    return;
  }
  if(tb->isEmpty()){
    QApplication::restoreOverrideCursor();
    return;
  }
  
  QString * q = new QString("CHECK TABLE ");
  q->append( *db );
  q->append( "." );
  q->append( *tb );
  
  slotDoQuery( db, q);
  
  QApplication::restoreOverrideCursor();
  
}

void SqlGuiDocMysql::slotTableOptimize( QString* db, QString* tb ){

  QApplication::setOverrideCursor( Qt::waitCursor );

  // std::cout << "Got Table-Optimize-Request..." << endl;

  if(db->isEmpty()){
    QApplication::restoreOverrideCursor();
    return;
  }
  if(tb->isEmpty()){
    QApplication::restoreOverrideCursor();
    return;
  }
  
  QString * q = new QString("OPTIMIZE TABLE ");
  q->append( *db );
  q->append( "." );
  q->append( *tb );
  
  slotDoQuery( db, q);
  
  QApplication::restoreOverrideCursor();
  
}

void SqlGuiDocMysql::slotTableRepair( QString* db, QString* tb ){

  QApplication::setOverrideCursor( Qt::waitCursor );

  // std::cout << "Got Table-Repair-Request..." << endl;

  if(db->isEmpty()){
    QApplication::restoreOverrideCursor();
    return;
  }
  if(tb->isEmpty()){
    QApplication::restoreOverrideCursor();
    return;
  }

  QString * q = new QString("REPAIR TABLE ");
  q->append( *db );
  q->append( "." );
  q->append( *tb );

  slotDoQuery( db, q);

  QApplication::restoreOverrideCursor();

}

void SqlGuiDocMysql::slotDoQuery(QString * Base, QString* Anfrage){

  //  // std::cout << "DB: " << Base->latin1() << endl;
  //  // std::cout << "Query: " << Anfrage->latin1() << endl;
  //  // std::cout << "Host: " << doc_host.latin1() << endl;
  //  // std::cout << "User: " << doc_uname.latin1() << endl;
  //  // std::cout << "PW: " << doc_pass.latin1() << endl;

  // return;

  MYSQL datenbank;
  mysql_init( &datenbank );
  MYSQL_RES * res;

  if(Base->isEmpty()){
    // std::cout << "No base selected..." << endl;
    QString *errmsg = new QString( kl->translate( i18n("No base selected.")));
    emit( DbQueryError( errmsg) );
    return;  
  }
  if(Anfrage->isEmpty()){
    // std::cout << "No SQL-String..." << endl;
    QString *errmsg = new QString( kl->translate( i18n("No SQL statement.")));
    emit( DbQueryError( errmsg) );
    return;  
  }

  // std::cout << "DB: " << Base->latin1() << endl;
  // std::cout << "SQL: " << Anfrage->latin1() << endl;


  if(!mysql_real_connect( &datenbank, doc_host.latin1() , doc_uname.latin1() , doc_pass.latin1(), "", doc_port, NULL, 0)){
    // std::cout << "Connection failed..." << endl;
    QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));

    doc_host.setLatin1("");
    doc_uname.setLatin1("");
    doc_pass.setLatin1("");

    emit( DbQueryError( errmsg) );
    mysql_close(&datenbank);
    return;
  }
  if(mysql_select_db( &datenbank, Base->latin1() )){
    // std::cout << "No such base..." << endl;
    QString *errmsg = new QString( kl->translate( i18n("No such database.")));
    emit( DbQueryError( errmsg) );
    mysql_close(&datenbank);
    return;
  }
  //
  // unter umstnden mu hier noch die escape-Funktion eingebaut werden
  //
  int okay = mysql_query( &datenbank, Anfrage->latin1() );
  if(! (res = mysql_use_result( &datenbank )) ){
    // std::cout << "Empty query..." << endl;
    ////    emit queryEmpty();
    if (okay){
      QString *errmsg = new QString( kl->translate( i18n("Query failed: ")));
      errmsg->append( mysql_error( &datenbank ) );
      emit( DbQueryError( errmsg) );
    } else {
      emit( QueryDone() );
    }
    if ( RESULT ){
      delete RESULT;
      RESULT = NULL;
    }
    if ( RESULT_Start ){
      delete RESULT_Start;
      RESULT_Start = NULL;
    }
    lastDB->setLatin1( Base->latin1() );
    //    emit( resultChanged() );
    //    emit( NewTableCreated() );
    //    lastDB->setLatin1( Base->latin1() );


    lastTable->setLatin1("");
    QUERY->setLatin1( Anfrage->latin1() );
    STATUS->setLatin1("");
    mysql_close(&datenbank);
    return;
  } else {
    // Jetzt die letzte Tabelle aus dem SQL-String raussuchen:
    // Wir brauchen uns nur um die "select"-Statements zu kmmern,
    // da nur diese eine Anzeige erzeugen, welche wir durch ein
    // "update"-statement modifizieren wollen.
      
    QString *sql = new QString( Anfrage->simplifyWhiteSpace() );
    QChar   bst = *(new QChar);
      
    lastDB->setLatin1( Base->latin1() );
      
    // std::cout << "Analysing SQL..." << sql->latin1() << endl;
      
    // hier mu evtl. noch nachgebessert werden, falls der tabellenname auf select endet...
    int select = sql->find("select ",0,false);
    
    if (select > -1){
      // Es handelt sich um ein select-statement:
      // std::cout << "Select-Statement..." << endl;
      // wir lschen die alte Variable:
      lastTable->setLatin1("");
      int from_pos = sql->find("from ",0,false);
      if (from_pos > -1){
	// Wir haben einen Tabellen-Namen nach dem From:
	// std::cout << "found from-expression... now looking for table..." << endl;
	from_pos += 5; // das wort "from " hat nochmals 5 buchstaben...
	int l = sql->length();
	bool weiter = true;
	int j = from_pos;
	while ( weiter )
	  {
            bst = sql->at(j);
            lastTable->append( bst );
            j++;
            if(j==l){weiter=false;}
            if(bst == " "){weiter=false;}
          }
        // std::cout << "tablename: " << lastTable->latin1() << endl;
      }
    }
  }
  
  //
  // Spaltennamen auslesen:
  //
  
  unsigned int num_fields;
  unsigned int r;
  column_t * columne;
  MYSQL_FIELD *fields;
  num_fields = mysql_num_fields( res );
  fields = mysql_fetch_fields( res );
  for(r = 0; r < num_fields; r++){
    columne = new column_t;
    column_t * column_alt;
    int p = strlen( fields[r].name );
    char * sp_name = new char[ p+1 ];
    if( p>0 ){
      sp_name = strcpy( sp_name, fields[r].name );
    }else{ 
      sp_name=NULL;
    }
    columne->name=sp_name;
    //    // std::cout << "Name: " << sp_name << endl;
    //    // std::cout << "Type: " << fields[r].type << endl;
    //    // std::cout << "Length: " << fields[r].length << endl;
    //    // std::cout << "Default: " << fields[r].def << endl;
    //    // std::cout << "Flags: " << fields[r].flags << endl;
    
    char * typus = new char[100];
    switch( fields[r].type){
    case FIELD_TYPE_TINY : {
      typus = strcpy(typus, "TINY");
      break;
    }
    case FIELD_TYPE_SHORT : {
      typus = strcpy(typus, "SHORT");
      break;
    }
    case FIELD_TYPE_LONG : {
      typus = strcpy(typus, "LONG");
      break;
    }
    case FIELD_TYPE_INT24 : {
      typus = strcpy(typus, "INT24");
      break;
    }
    case FIELD_TYPE_LONGLONG : {
      typus = strcpy(typus, "LONGLONG");
      break;
    }
    case FIELD_TYPE_DECIMAL : {
      typus = strcpy(typus, "DECIMAL");
      break;
    }
    case FIELD_TYPE_FLOAT : {
      typus = strcpy(typus, "FLOAT");
      break;
    }
    case FIELD_TYPE_DOUBLE : {
      typus = strcpy(typus, "DOUBLE");
      break;
    }
    case FIELD_TYPE_TIMESTAMP : {
      typus = strcpy(typus, "TIMESTAMP");
      break;
    }
    case FIELD_TYPE_DATE : {
      typus = strcpy(typus, "DATE");
      break;
    }
    case FIELD_TYPE_TIME : {
      typus = strcpy(typus, "TIME");
      break;
    }
    case FIELD_TYPE_DATETIME : {
      typus = strcpy(typus, "DATETIME");
      break;
    }
    case FIELD_TYPE_YEAR : {
      typus = strcpy(typus, "YEAR");
      break;
    }
    case FIELD_TYPE_STRING : {
      typus = strcpy(typus, "STRING");
      break;
    }
    case FIELD_TYPE_BLOB : {
      typus = strcpy(typus, "BLOB");
      break;
    }
    case FIELD_TYPE_SET : {
      typus = strcpy(typus, "SET");
      break;
    }
    case FIELD_TYPE_ENUM : {
      typus = strcpy(typus, "ENUM");
      break;
    }
    case FIELD_TYPE_NULL : {
      typus = strcpy(typus, "NULL");
      break;
    }
    case FIELD_TYPE_NEWDATE : {
      typus = strcpy(typus, "NEWDATE");
      break;
    }
    case FIELD_TYPE_TINY_BLOB : {
      typus = strcpy(typus, "TINY_BLOB");
      break;
    }
    case FIELD_TYPE_MEDIUM_BLOB : {
      typus = strcpy(typus, "MEDIUM_BLOB");
      break;
    }
    case FIELD_TYPE_LONG_BLOB : {
      typus = strcpy(typus, "LONG_BLOB");
      break;
    }
    case FIELD_TYPE_VAR_STRING : {
      typus = strcpy(typus, "VAR_STRING");
      break;
    }
    }
    columne->type = typus;
    
    if(r==0) lastCols=columne;
    else column_alt->next=columne;
    column_alt=columne;
  }
  if( columne ) columne->next=NULL;
  
  // Datenstze auslesen:
  MYSQL_ROW row;
  unsigned int num_cols = mysql_field_count( &datenbank );
  
  if ( RESULT ){
    delete RESULT;
    RESULT = NULL;
  }
  if ( RESULT_Start ){
    clearResult();
    delete RESULT_Start;
    RESULT_Start = NULL;
  }
  
  row_t * hilf_row = NULL;
  
  while( (row = mysql_fetch_row( res )) ){
    
    row_t * rw = new row_t;
    rw->next = NULL;
    
    unsigned long * len = mysql_fetch_lengths(res);
    
    field_t * hilf_col = NULL;
    
    // alle Elemente der Reihe nach anhngen:
    for( unsigned int i = 0; i<num_cols; i++){
      char * rw_name = new char[ len[ i ]+1 ];
      field_t * col = new field_t;
      if( len[i] )strcpy( rw_name, row[i] );
      else rw_name=NULL;
      col->value = rw_name;
      if( hilf_col ){
        hilf_col->next = col;
      } else {
        rw->feld = col;
      }
      hilf_col = col;
    }
    // Abbruchbedingung:
    if(hilf_col)hilf_col->next=NULL;
    // nchste Reihe:
    if (hilf_row){
      hilf_row->next = rw;
    } else {
      RESULT_Start = rw;
    }
    hilf_row = rw;
  }
  // Abbruchbedingung:
  if(hilf_row)hilf_row->next=NULL;

  mysql_close( &datenbank );
  emit( resultChanged() );

  // Also lesen wir die Sachen aus den bereits eingescannten Daten aus...

  column_t * sp;
  table_t   * tab;
  base_t    * bse = DB;
  while( bse ){
    if ( strcmp(bse->name, lastDB->latin1() ) == 0 ){
      tab = bse->table_list;
      // // std::cout << "Found last db in DB-Object " << endl;
      bse = NULL;
      while( tab ){
        if ( strcmp( tab->name, lastTable->latin1() ) == 0 ){	
          sp = tab->column_list;
          // // std::cout << "Found last table in DB-Object " << endl;
          tab = NULL;
          while(sp){
            if ( sp->pri == 'Y' ){
              // // std::cout << "Found primary-column in last table " << endl;
              primKey->setLatin1( sp->name );
              // // std::cout << primKey->latin1() << endl;
              sp = NULL;
            } else {
              sp = sp->next;
            }
          }
        } else {	
          tab = tab->next;
        }
      }	
    } else {
      bse = bse->next;
    }
  }
  emit lastActionChanged( lastDB, lastTable, primKey);
}

bool SqlGuiDocMysql::openDocument(const QString& url, const QString& uname, const QString& pwd, const QString& base){
  const char * host;
  const char * user;
  const char * pass;
  //char * info;
  int anz_db  = 0;
  int anz_tab = 0;

  connectionEnabled=false;

  QObject * myApp = parent();
  QApplication::setOverrideCursor( Qt::WaitCursor );
  
  if ( ! url.isEmpty() ){
    
    // std::cout << "Aufruf von openDocument" << endl;
    
    host = url.latin1();
    user = uname.latin1();
    pass = pwd.latin1();
    
    char * portpart = strchr(host, ':');
    if(portpart != NULL){
      doc_port = atoi(portpart+1);
      *portpart = 0;
      // std::cout << "host: " << host << endl << "port: " << doc_port << endl;
    }else{
      // std::cout << "no port: " << host << endl;
    }
    
    doc_host.setLatin1(host);
    doc_uname.setLatin1(user);
    doc_pass.setLatin1(pass);
    
    unsigned int num_rows;
    unsigned int i;
    
    MYSQL datenbank;
    mysql_init( &datenbank );
    if(! &datenbank ){
      // std::cout << "ERROR initializing MYSQL-Object... :-(" << endl;
      delete DB;
      DB=NULL;
    } else{
      mysql_real_connect(&datenbank, host , user , pass, "", doc_port, NULL, 0);
      if(! &datenbank ){
        // std::cout << "ERROR connecting to server... :-((" << endl;
        QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
        QApplication::restoreOverrideCursor();

        doc_host.setLatin1("");
        doc_uname.setLatin1("");
        doc_pass.setLatin1("");

        emit( DbQueryError( errmsg) );
        delete DB;
        DB=NULL;
      } else {

        const char * info = mysql_get_server_info(&datenbank);
        // std::cout << "MySQL-Info: " << info << endl;
        doc_url.setFileName( kl->translate( i18n("Connected")));

        MYSQL_RES * result;
        result = mysql_list_dbs( &datenbank, "%");

        if ( ! result ){
          // std::cout << "ERROR getting DBs... :-(" << endl;
          QString *errmsg = new QString( kl->translate( i18n("Error getting databases...\nMaybe the server is busy or down...")));
          QApplication::restoreOverrideCursor();
          emit( DbQueryError( errmsg) );
          emit( loginError() );
          delete DB;
          DB=NULL;
          return false;
        } else {
  
          //  num_rows = mysql_num_rows(result);
          //  // std::cout << "Got " << num_rows << " DBs..." << endl;
  
          MYSQL_ROW row;
  
          clearResult();
          DB=NULL;
          base_t * hilf=NULL;
  
          while( (row = mysql_fetch_row(result)) ){
            //	    // std::cout << row[0] << " ";
            // Fr jede DB erzeugen wir einen neuen eintrag im Objekt DB:
            base_t * temp = new base_t;
            char * db_name=NULL;
            db_name = new char[ strlen(row[0])+1 ];
            db_name = strcpy( db_name , row[0]);
            temp->name = db_name;
            // eigentlich knnten wir hier auch schon alle Tabellen fr die jeweilige DB auslesen,
            // aber das gibt Probleme bei den verschachtelten MySQL-Aufrufen... deswegen lesen
            // wir die Tabellen spter separat aus...
            if (DB==NULL){
              DB = temp;
            } else {
              hilf->next = temp;
            }
	    hilf = temp;
          }
          hilf->next=NULL;
          //	  // std::cout << endl;
          mysql_free_result( result );
        }
	
        mysql_close( &datenbank );
	// std::cout << "Connection is closed..." << endl;
      }
    }
    
    // Jetzt die Tabellen auslesen:
    mysql_init( &datenbank );
    MYSQL_RES * dbs;
    if(! &datenbank ){
      // std::cout << "ERROR initializing MYSQL-Object... :-(" << endl;
      delete DB;
      DB=NULL;
    } else{
      MYSQL* datenbank_p = mysql_real_connect(&datenbank, host , user , pass, "", doc_port, NULL, 0);
      if(! datenbank_p ){
        // std::cout << "ERROR connecting to server... :-((" << endl;
        QString *errmsg = new QString(  kl->translate( i18n("Error connecting to server.")));
        QApplication::restoreOverrideCursor();
        doc_host.setLatin1("");
        doc_uname.setLatin1("");
        doc_pass.setLatin1("");
        emit( DbQueryError( errmsg) );
        delete DB;
        DB=NULL;
      } else {
        base_t * test = DB;
        // // std::cout << "Lese Tabellen aus..." << endl;
        while( test ){
          int okay = mysql_select_db( &datenbank, test->name);
          // // std::cout << endl << "Selected " << test->name << ", Status: " << okay << endl;
          dbs = NULL;
          if(okay==0){  dbs = mysql_list_tables( &datenbank, "%");}
          // // std::cout << mysql_error( &datenbank );
          // // std::cout << mysql_error( &datenbank );
          // unsigned int num_tabs = mysql_num_rows( dbs );
          // // std::cout << "DB: " << test->name << ": "<< "Got " << num_tabs << " Tables..." << endl;
  
          MYSQL_ROW tables;
          //	  // std::cout << "Tabellen eingelesen..." << endl;
          table_t * alt = NULL;
          test->table_list = NULL;
          if(dbs){
            while( (tables = mysql_fetch_row( dbs )) ){
              //	    // std::cout << tables[0] << " ";
      
              char * tab_name = new char[ strlen( tables[0] ) + 1 ];
              strcpy(tab_name , tables[0] );
      
              table_t* test3 = new table_t;
              test3->name = tab_name;
      
              if(alt){
                alt->next=test3;
              } else {
                test->table_list = test3;
              }
              alt = test3;
            }
          }
          if(alt)alt->next=NULL;
          if(dbs){mysql_free_result( dbs );}
          if(test)test = test->next;
  
        }
        mysql_close(&datenbank);
        mysql_close( & datenbank );
      }
    }
    
    
    // // std::cout << "Okay, I got the tables..." << endl;
    // // std::cout << "Now getting the columns..." << endl;
    
    mysql_init( &datenbank );
    MYSQL_RES * cols;
    if(! &datenbank ){
      // std::cout << "ERROR initializing MYSQL-Object... :-(" << endl;
      delete DB;
      DB=NULL;
    } else{
      MYSQL* datenbank_p = mysql_real_connect(&datenbank, host , user , pass, "", doc_port, NULL, 0);
      if(! datenbank_p ){
        // std::cout << "ERROR connecting to server... :-((" << endl;
        QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
        QApplication::restoreOverrideCursor();
        doc_host.setLatin1("");
        doc_uname.setLatin1("");
        doc_pass.setLatin1("");
        emit( DbQueryError( errmsg) );
        delete DB;
        DB=NULL;
      } else {

        //
        // Kann jemand mal dieses dumme Api dazu bringen, verschachtelte Abfragen durchzufhren?
        //
        // etwa so:
        //
        // abfrage1 = list_dbs();
        // while ( db = fetch_row( abfrage1) ){
        //   abfrage2 = list_tables( db );
        //   while( tb = fetch_row( abfrage2 ) ){
        //     abfrage3 = list_fields( tb );
        //     while( fd = fetch_row( abfrage3 ) ){
        //       // std::cout << bla-bla...
        //     }
        //   }
        // }
        //
        // Welche Rolle spielen dabei use_result und store_result?
        //

        // Jetzt die Spalten der Tabellen raussuchen...
        //
        base_t  * bse = DB;
        table_t * tbl = bse->table_list;
        while(bse){
          tbl = bse->table_list;
          while( tbl ){
            MYSQL_ROW columns;
            char * query = new char[ 20 + strlen(tbl->name) ];
            int okay = mysql_select_db( &datenbank, bse->name);
            // std::cout << mysql_error( &datenbank );
            query = strcpy( query, "SHOW COLUMNS FROM ");
            query = strcat(query, tbl->name );
    
            okay = mysql_query( &datenbank, query );
            cols = mysql_use_result( &datenbank );
            //	    // std::cout << "DB: " << bse->name  << endl << query << endl << mysql_error( &datenbank ) << endl;
            //	    // std::cout << "Selected Table: " << tbl->name << endl;
            column_t * alt2 = NULL;
            while( (columns = mysql_fetch_row( cols ) ) ){
              char * sp_name = new char[ strlen( columns[0] ) +1 ];
              strcpy( sp_name, columns[0] );
              char * sp_type = new char[ strlen( columns[1] ) +1 ];
              strcpy( sp_type, columns[1] );
              char * sp_extra = new char[ strlen( columns[5] ) +1 ];
              strcpy( sp_extra, columns[5] );
              //	      // std::cout << columns[0] << " " << columns[1] << " " << columns[2] << endl;
              column_t * spl = new column_t;
              spl->name = sp_name;
              spl->type = sp_type;
              spl->extra_flags = sp_extra;
              if( (columns[3][0] == 'P') && (columns[3][1] == 'R') && (columns[3][2] == 'I') ){
                spl->pri='Y';
              }else{
                spl->pri=' ';
              }
      
              spl->next=NULL;
              //	      // std::cout << sp_name << " ";
              if( alt2 ){
                alt2->next = spl;
              } else {
                tbl->column_list = spl;
              }
              alt2 = spl;
            }
            mysql_free_result( cols );
            if(tbl)tbl = tbl->next;
          }
          if(bse)bse=bse->next;
        }
      }
      mysql_close(&datenbank);
    }    
  }
  //// std::cout << "Okay... now emmitting  connectionChanged()" << endl;

  emit( connectionChanged() );
  
  if ( RESULT ){
    delete RESULT;
    RESULT = NULL;
  }
  if ( RESULT_Start ){
    delete RESULT_Start;
    RESULT_Start = NULL;
  }
  
  if(DB){
    basePosition = DB;
    tablePosition = DB->table_list;
    if (tablePosition) {
      columnPosition = tablePosition->column_list;
    }else{
      columnPosition = NULL;
    }
    rowPosition = NULL;      // not needed (implemented) now
    fieldPosition = NULL;    // not needed (implemented) now
  } else {
    basePosition = NULL;
    tablePosition = NULL;
    columnPosition = NULL;
    rowPosition = NULL;      // not needed (implemented) now
    fieldPosition = NULL;    // not needed (implemented) now
  }
  
  // emit( resultChanged() );
  emit loginComplete();
  connectionEnabled=true;
  if (! DB ){
    connectionEnabled=false;
    emit loginError();
  }
  QApplication::restoreOverrideCursor();
  return true;
}

extern "C" SqlGuiDoc* create(QWidget *parent, const char *name){
  return new SqlGuiDocMysql( parent, name );
}

extern "C" void destroy( SqlGuiDoc* p){
  delete p;
}

//#include "sqlguidoc_mysql.moc.cpp"
